------------------------------------------------------------------------------
--                                                                          --
--                            Matreshka Project                             --
--                                                                          --
--                          Ada Modeling Framework                          --
--                                                                          --
--                        Runtime Library Component                         --
--                                                                          --
------------------------------------------------------------------------------
--                                                                          --
-- Copyright © 2011-2012, Vadim Godunko <vgodunko@gmail.com>                --
-- All rights reserved.                                                     --
--                                                                          --
-- Redistribution and use in source and binary forms, with or without       --
-- modification, are permitted provided that the following conditions       --
-- are met:                                                                 --
--                                                                          --
--  * Redistributions of source code must retain the above copyright        --
--    notice, this list of conditions and the following disclaimer.         --
--                                                                          --
--  * Redistributions in binary form must reproduce the above copyright     --
--    notice, this list of conditions and the following disclaimer in the   --
--    documentation and/or other materials provided with the distribution.  --
--                                                                          --
--  * Neither the name of the Vadim Godunko, IE nor the names of its        --
--    contributors may be used to endorse or promote products derived from  --
--    this software without specific prior written permission.              --
--                                                                          --
-- THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS      --
-- "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT        --
-- LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR    --
-- A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT     --
-- HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,   --
-- SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED --
-- TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR   --
-- PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF   --
-- LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING     --
-- NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS       --
-- SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.             --
--                                                                          --
------------------------------------------------------------------------------
--  $Revision: 2580 $ $Date: 2012-03-11 12:40:11 +0100 (Sun, 11 Mar 2012) $
------------------------------------------------------------------------------
--  Root package of visitor pattern implementation. It provides two interface
--  types:
--
--   - Abstract_Visitor - root interface type for metamodel specific visitors.
--     Derived types provides pair of operations for each non-abstract class of
--     elements: Enter_<Class> and Leave_<Class>. They are called by iterator
--     when it enter or leave element;
--
--   - Abstract_Iterator - root interface type for metamodel specific
--     iterators. Derived types provides one operation for each non-abstract
--     class of elements: Visit_<Class>. It is responsible to call
--     Enter_<Class> and Leave_<Class> operations of visitor as well as to
--     control iteration.
------------------------------------------------------------------------------
with AMF.Elements;

package AMF.Visitors is

   pragma Preelaborate;

   type Traverse_Control is
    (Continue,
     Abandon_Children,
     Abandon_Sibling,
     Terminate_Immediately);

   type Abstract_Visitor is limited interface;
   --  Root interface type for metamodel specific visitors.

   type Abstract_Iterator is limited interface;
   --  Root interface type for metamodel specific iterators.

   procedure Visit
    (Self    : in out Abstract_Iterator'Class;
     Visitor : in out Abstract_Visitor'Class;
     Element : not null AMF.Elements.Element_Access;
     Control : in out Traverse_Control);
   --  Visit specified element by calling:
   --
   --   - Enter_<Class> operation on visitor;
   --
   --   - Visit_<Class> operation on iterator;
   --
   --   - Leave_<Class> operation on visitor.
   --
   --  Internal double dispatching mechanism is used by implementation, so it
   --  works on elements of any metamodel.

end AMF.Visitors;
